import structio
from structio.core.task import Task
from structio.core.run import current_loop
from typing import Callable, Coroutine, Any
from structio.core.time.queue import TimeQueue
from structio.core.syscalls import suspend, checkpoint
from structio.exceptions import Cancelled, StructIOException


class TaskScope:
    """
    A task scope
    """

    def __init__(
        self,
        timeout: int | float | None = None,
        silent: bool = False,
        shielded: bool = False,
    ):
        # When do we expire?
        self._timeout = timeout or float("inf")
        # This is updated with the actual wall clock
        # time of when we expire each time that the
        # timeout is modified
        self.deadline = -1
        # Do we raise an error on timeout?
        self.silent = silent
        # Has a cancellation attempt been done?
        self.attempted_cancel: bool = False
        # Have we been cancelled?
        self.cancelled: bool = False
        # Have we timed out?
        self.timed_out: bool = False
        # Can we be indirectly cancelled? Note that this
        # does not affect explicit cancellations via the
        # cancel() method
        self.shielded: bool = shielded
        # Data about inner and outer scopes.
        # This is used internally to make sure
        # nesting task scopes works as expected
        self.inner: list[TaskScope] = []
        self.outer: TaskScope | None = None
        # Which tasks do we contain?
        self.tasks: list[Task] = []
        self.owner: Task | None = None

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        self._timeout = value
        self.deadline = current_loop().clock.deadline(self.timeout)

    def cancel(self):
        """
        Cancels the task scope and all the work
        that belongs to it
        """

        current_loop().cancel_scope(self)

    def get_effective_deadline(self) -> tuple[float, "TaskScope"]:
        """
        Returns the effective deadline of the whole
        cancel scope
        """

        queue = TimeQueue()
        if self.shielded:
            return float("inf"), self
        queue.put(self, self.deadline)
        for child in self.children:
            if child.shielded:
                return float("inf"), self
            deadline, scope = child.get_effective_deadline()
            queue.put(scope, deadline)
        return queue.get_closest_deadline(), queue.get()[0]

    def __repr__(self):
        return f"{self.__class__.__name__}(owner={self.owner}, timeout={self.timeout})"

    def __enter__(self):
        current_loop().init_scope(self)
        return self

    def __exit__(self, exc_type: type, exc_val: BaseException, exc_tb):
        current_loop().close_scope(self)
        if exc_val and isinstance(exc_val, structio.TimedOut):
            if exc_val.scope is self:
                return self.silent

    # Just a recursive helper
    def _get_children(self, lst=None):
        if lst is None:
            lst = []
        for child in self.inner:
            lst.append(child)
            child._get_children(lst)
        return lst

    @property
    def children(self) -> list["TaskScope"]:
        """
        Gets all the scopes contained within this one
        """

        return self._get_children()

    def done(self):
        """
        Returns whether the task scope has finished executing
        """

        if not all(child.done() for child in self.children):
            return False
        return all(task.done() for task in self.tasks)


class TaskPool:
    """
    A task pool
    """

    def __init__(self):
        self.entry_point: Task | None = None
        self.scope: TaskScope = TaskScope(timeout=float("inf"))
        # This pool's parent
        self.outer: TaskPool | None = None
        # Have we errored out?
        self.error: BaseException | None = None
        # Have we exited? This is so we can forbid reuse of
        # dead task pools
        self._closed: bool = False

    async def __aenter__(self):
        if self._closed:
            raise StructIOException("task pool is closed")
        self.scope.__enter__()
        current_loop().init_pool(self)
        return self

    async def __aexit__(self, exc_type: type, exc_val: BaseException, exc_tb):
        try:
            if exc_val:
                raise exc_val.with_traceback(exc_tb)
            elif not self.done():
                await suspend()
                assert self.done()
            else:
                await checkpoint()
        except Cancelled as e:
            self.error = e
            self.scope.cancelled = True
        except (Exception, KeyboardInterrupt) as e:
            self.error = e
            self.scope.cancel()
        finally:
            self.scope.__exit__(exc_type, exc_val, exc_tb)
            current_loop().close_pool(self)
            self._closed = True
            if self.error:
                raise self.error

    def done(self):
        """
        Returns whether the task pool's internal
        task scope has finished executing. Note
        that this does not take the scope's owner
        into account!
        """

        return self.scope.done()

    def spawn(
        self, func: Callable[[Any, Any], Coroutine[Any, Any, Any]], *args
    ) -> Task:
        """
        Schedule a new concurrent task for execution in the task pool from the given
        async function. All positional arguments are passed to the underlying coroutine
        (for keyword arguments, consider using `functools.partial`). A Task object is
        returned. Note that the coroutine is merely scheduled to run and does not begin
        executing until it is picked by the scheduler later on
        """

        if self._closed:
            raise StructIOException("task pool is closed")

        return current_loop().spawn(func, *args)
