import traceback
import warnings
from types import FrameType
from structio.abc import (
    Kernel,
    Clock,
    Debugger,
    IOManager,
    SignalManager,
    SchedulingPolicy,
)
from structio.io import FdWrapper
from structio.core.context import TaskPool, TaskScope
from structio.core.task import Task, TaskState
from structio.util.ki import CTRLC_PROTECTION_ENABLED
from structio.exceptions import (
    StructIOException,
    Cancelled,
    TimedOut,
    ResourceClosed,
    ResourceBroken,
)
from collections import deque
from typing import Callable, Coroutine, Any
from itertools import count
from functools import partial
import signal
import sniffio


class DefaultKernel(Kernel):
    """
    An asynchronous event loop implementation
    supporting generic scheduling policies
    """

    def __init__(
        self,
        policy: SchedulingPolicy,
        clock: Clock,
        io_manager: IOManager,
        signal_managers: list[SignalManager],
        tools: list[Debugger] | None = None,
        restrict_ki_to_checkpoints: bool = False,
    ):
        super().__init__(
            policy,
            clock,
            io_manager,
            signal_managers,
            tools,
            restrict_ki_to_checkpoints,
        )
        # Tasks that are run when the entry point
        # exits correctly
        self.shutdown_tasks: deque[
            tuple[Any, Callable[[Any, Any], Coroutine[Any, Any, Any]], tuple[Any, ...]]
        ] = deque()
        # Internal identifier generator for shutdown tasks
        self._shutdown_task_ident = count(0)
        # Data to send back to tasks
        self.data: dict[Task, Any] = {}
        # Have we handled SIGINT?
        self._sigint_handled: bool = False
        # Pool where the entry point and system tasks run
        self.pool = TaskPool()
        self.current_scope = self.pool.scope
        self._shutting_down = False

    def get_closest_deadline(self) -> Any:
        return self.policy.get_closest_deadline()

    def wait_readable(self, resource: FdWrapper):
        self.current_task: Task
        self.io_manager.request_read(resource, self.current_task)

    def wait_writable(self, resource: FdWrapper):
        self.current_task: Task
        self.io_manager.request_write(resource, self.current_task)

    def notify_closing(
        self, resource: FdWrapper, broken: bool = False, owner: Task | None = None
    ):
        if not broken:
            exc = ResourceClosed("stream has been closed")
        else:
            exc = ResourceBroken("stream might be corrupted")
        owner = owner or self.current_task
        reader = self.io_manager.get_reader(resource)
        writer = self.io_manager.get_writer(resource)
        if reader and reader is not owner:
            self.throw(reader, exc)
        if writer and writer is not owner:
            self.throw(writer, exc)
        self.reschedule_running()

    def event(self, evt_name: str, *args):
        if not callable(getattr(Debugger, evt_name, None)):
            warnings.warn(f"Invalid debugging event fired: {evt_name!r}")
            return
        for tool in self.tools:
            # Not all tools must implement all debugging events!
            if f := getattr(tool, evt_name, None):
                try:
                    f(*args)
                except BaseException as e:
                    # We really can't afford to have our internals explode,
                    # sorry!
                    warnings.warn(
                        f"Exception during debugging event delivery in {f!r} ({evt_name!r}): {type(e).__name__} -> {e}",
                    )
                    traceback.print_tb(e.__traceback__)
                    # We disable the tool, so it can't raise at the next debugging
                    # event
                    self.tools.remove(tool)

    def done(self):
        if self.entry_point.done() and not self._shutting_down:
            return True
        if any(
            [
                self.policy.has_next_task(),
                self.policy.has_paused_task(),
                self.io_manager.pending(),
            ]
        ):
            return False
        if not self.pool.done():
            return False
        return True

    def spawn(
        self,
        func: Callable[[Any, Any], Coroutine[Any, Any, Any]],
        *args,
        ki_protected: bool = False,
        pool: TaskPool = None,
        system_task: bool = False,
        entry_point: bool = False
    ):
        if isinstance(func, partial):
            name = func.func.__name__ or repr(func.func)
        else:
            name = func.__name__ or repr(func)
        if pool is None:
            if system_task:
                pool = self.pool
            else:
                pool = self.current_pool
        if pool is self.pool and not entry_point:
            system_task = True
        task = Task(name, func(*args), pool.scope, pool, is_system_task=system_task)
        pool.scope.tasks.append(task)
        # We inject our magic secret variable into the coroutine's stack frame, so
        # we can look it up later
        task.coroutine.cr_frame.f_locals.setdefault(
            CTRLC_PROTECTION_ENABLED, ki_protected
        )
        self.policy.schedule(task)
        self.event("on_task_spawn", task)
        return task

    def spawn_system_task(
        self, func: Callable[[Any, Any], Coroutine[Any, Any, Any]], *args
    ):
        return self.spawn(func, *args, ki_protected=True, system_task=True)

    def signal_notify(self, sig: int, frame: FrameType):
        match sig:
            case signal.SIGINT:
                if self.restrict_ki_to_checkpoints:
                    self._sigint_handled = True
                    # Poke the event loop with a stick ;)
                    self.policy.schedule(self._pick_ki_task())
                else:
                    self.raise_ki()
            case _:
                pass

    def step(self):
        """
        Run a single task step (i.e. until an "await" to our
        primitives somewhere)
        """

        if not self.policy.has_next_task():
            return
        self.current_task = self.policy.get_next_task()
        self.current_task: Task
        while self.current_task.done():
            if not self.policy.has_next_task():
                return
            self.current_task = self.policy.get_next_task()
        self.current_task: Task
        if self.current_task.done():
            return
        runner = partial(
            self.current_task.coroutine.send, self.data.pop(self.current_task, None)
        )
        if self.current_task.pending_cancellation:
            self.cancel_task(self.current_task)
            return
        elif self._sigint_handled:
            self.raise_ki(self.current_task)
        self.event("before_task_step", self.current_task)
        self.current_task.state = TaskState.RUNNING
        self.current_task.paused_when = 0
        self.current_pool = self.current_task.pool
        self.current_scope = self.current_task.scope
        data = self.handle(runner, self.current_task)
        self._sigint_handled = False
        if data is not None:
            method, args, kwargs = data
            self.suspend()
            if not callable(getattr(self, method, None)):
                # This if block is meant to be triggered by other async
                # libraries, which most likely have different method names and behaviors
                # compared to us. If you get this exception, and you're 100% sure you're
                # not mixing async primitives from other libraries, then it's a bug!
                self.throw(
                    self.current_task,
                    StructIOException(
                        "Uh oh! Something bad just happened: did you try to mix "
                        "primitives from other async libraries?"
                    ),
                )
            # Sneaky method call, thanks to David Beazley for this ;)
            getattr(self, method)(*args, **kwargs)
        self.event("after_task_step", self.current_task)

    def throw(self, task: Task, err: BaseException):
        if task.done():
            return
        self.handle(partial(task.coroutine.throw, err), task)

    def reschedule(self, task: Task):
        if task.done():
            return
        self.policy.schedule(task)

    def check_cancelled(self, schedule: bool = True):
        if self.current_task.pending_cancellation:
            self.current_task: Task
            self.cancel_task(self.current_task)
        elif schedule:
            self.current_task: Task
            # We reschedule the caller immediately!
            self.policy.schedule(self.current_task)

    def schedule_point(self):
        self.reschedule_running()

    def sleep(self, amount):
        """
        Puts the current task to sleep for the given amount of
        time as defined by our current clock
        """

        # Just to avoid code duplication, you know
        self.suspend()
        if amount > 0:
            self.event("before_sleep", self.current_task, amount)
            self.current_task.next_deadline = self.clock.deadline(amount)
            self.current_task: Task
            self.policy.pause(self.current_task)
        else:
            # If sleep is called with 0 as argument,
            # then it's just a checkpoint!
            self.schedule_point()
            self.check_cancelled(schedule=False)

    def check_scopes(self):
        expired = set()
        for scope in self.pool.scope.children:
            deadline, actual = scope.get_effective_deadline()
            if deadline <= self.clock.current_time() and not actual.timed_out:
                expired.add(actual)
        for scope in expired:
            scope.timed_out = True
            error = TimedOut("timed out")
            error.scope = scope
            scope.cancel()
            self.throw(scope.owner, error)
            if not self.policy.is_scheduled(scope.owner):
                self.reschedule(scope.owner)

    def wakeup(self):
        while (
            self.policy.has_paused_task()
            and self.policy.peek_paused_task().next_deadline
            <= self.clock.current_time()
        ):
            task = self.policy.get_paused_task()
            task.next_deadline = 0
            self.event(
                "after_sleep", task, task.paused_when - self.clock.current_time()
            )
            self.reschedule(task)

    def _pick_ki_task(self):
        """
        Attempts to find a suitable task to
        throw KeyboardInterrupt into
        """

        if self.policy.has_next_task() and not self.policy.peek_next_task().is_system_task:
            return self.policy.get_next_task()
        if self.policy.has_paused_task():
            return self.policy.get_paused_task()
        if not self.entry_point.done():
            return self.entry_point
        raise StructIOException("unable to find a task to throw KeyboardInterrupt into")

    def raise_ki(self, task: Task | None = None):
        """
        Raises a KeyboardInterrupt exception into a
        task: If one is passed explicitly, the exception
        is thrown there, otherwise a suitable task is
        awakened and thrown into
        """

        self.throw(task or self._pick_ki_task(), KeyboardInterrupt())
        self._sigint_handled = False

    def _tick(self):
        """
        Runs a single event loop tick
        """

        self.wakeup()
        self.check_scopes()
        self.step()
        if self.io_manager.pending():
            self.io_manager.wait_io()

    def run(self):
        """
        This is the actual "loop" part
        of the "event loop"
        """

        while not self.done():
            self._tick()
        self._shutting_down = True
        # Ensure all system tasks have a chance to spin up
        while any(
            task.is_system_task and task.state == TaskState.INIT
            for task in self.pool.scope.tasks
        ):
            self._tick()
        # Cancel the system pool and wait for cancellation
        # to be delivered
        self.pool.scope.cancel()
        while not self.done():
            self._tick()
        # Sanity checking
        assert self.pool.scope.attempted_cancel
        assert self.pool.scope.cancelled
        assert self.pool.done()
        # Reset some stuff. Should probably initialize
        # a new task scope, but I'm lazy and this works
        self.pool.scope.attempted_cancel = False
        self.pool.scope.cancelled = False
        if self.entry_point.state == TaskState.FINISHED:
            while True:
                # Spawn all the shutdown tasks that are currently registered
                while self.shutdown_tasks:
                    # Shutdown tasks can be added even while the
                    # event loop is shutting down!
                    _, func, args = self.shutdown_tasks.popleft()
                    self.spawn_system_task(func, *args)
                # Run all the shutdown tasks to completion
                while not self.done():
                    self._tick()
                # Were more shutdown tasks scheduled? If not, we're done.
                # Otherwise, we'll run them at the next iteration
                if not self.shutdown_tasks:
                    break

        self.io_manager.close()
        self.close()

    def reschedule_running(self):
        """
        Reschedules the currently running task
        """

        self.current_task: Task
        self.reschedule(self.current_task)

    def handle(self, func: Callable, task: Task):
        """
        Convenience method for handling various exceptions
        from tasks
        """

        old_name, sniffio.thread_local.name = sniffio.thread_local.name, "structured-io"
        try:
            return func()
        except StopIteration as ret:
            # At the end of the day, coroutines are generator functions with
            # some tricky behaviors, and this is one of them. When a coroutine
            # hits a return statement (either explicit or implicit), it raises
            # a StopIteration exception, which has an attribute named value that
            # represents the return value of the coroutine, if it has one. Of course
            # this exception is not an error, and we should happily keep going after it:
            # most of this code below is just useful for internal/debugging purposes
            task.state = TaskState.FINISHED
            task.result = ret.value
            self.on_success(task)
        except Cancelled:
            # When a task needs to be cancelled, we try to do it gracefully first:
            # if the task is paused in either I/O or sleeping, that's perfect.
            # But we also need to cancel a task if it was not sleeping or waiting on
            # any I/O because it could never do so (therefore blocking everything
            # forever). So, when cancellation can't be done right away, we schedule
            # it for the next execution step of the task. We will also make sure
            # to re-raise cancellations at every checkpoint until the task lets the
            # exception propagate into us, because we *really* want the task to be
            # cancelled
            task.state = TaskState.CANCELLED
            task.pending_cancellation = False
            self.on_cancel(task)
        except (Exception, KeyboardInterrupt) as err:
            # Any other exception is caught here
            task.exc = err
            err.scope = task.scope
            task.state = TaskState.CRASHED
            self.on_error(task)
        finally:
            sniffio.thread_local.name = old_name

    def release_resource(self, resource: FdWrapper):
        self.io_manager.release(resource)
        self.reschedule_running()

    def release(self, task: Task):
        """
        Releases the timeouts and associated
        I/O resources that the given task owns
        """

        self.io_manager.release_task(task)
        self.policy.discard(task)

    def _reschedule_scope_tree(self, scope: TaskScope):
        # Walk up the scope tree and reschedule all necessary
        # tasks
        while scope and scope.done() and scope is not self.pool.scope:
            self.reschedule(scope.owner)
            scope = scope.outer

    def on_success(self, task: Task):
        """
        The given task has exited gracefully: hooray!
        """

        assert task.state == TaskState.FINISHED
        self._reschedule_scope_tree(task.scope)
        self.event("on_task_exit", task)
        self.release(task)

    def close(self, force: bool = False):
        super().close(force)
        # Cancel *ALL* tasks
        self.pool.scope.cancel()

    def on_error(self, task: Task):
        """
        The given task raised an exception
        """

        assert task.state == TaskState.CRASHED
        self.event("on_exception_raised", task, task.exc)
        if task.is_system_task:
            self.close(force=True)
            raise task.exc from StructIOException(f"system task {task} crashed")
        scope = task.scope
        self._reschedule_scope_tree(scope)
        self.release(task)
        self.cancel_scope(scope)
        if task is not scope.owner and scope.done():
            self.throw(scope.owner, task.exc)

    def on_cancel(self, task: Task):
        """
        The given task crashed because of a
        cancellation exception
        """

        assert task.state == TaskState.CANCELLED
        self._reschedule_scope_tree(task.scope)
        self.event("after_cancel", task)
        self.release(task)

    def init_scope(self, scope: TaskScope):
        self.current_task.scope = scope
        scope.deadline = self.clock.deadline(scope.timeout)
        scope.owner = self.current_task
        self.current_scope.inner.append(scope)
        scope.outer = self.current_scope
        self.current_scope = scope

    def close_scope(self, scope: TaskScope):
        self.current_scope = scope.outer
        self.current_scope.inner = []
        self.current_task.scope = self.current_scope

    def cancel_task(self, task: Task):
        if task.done():
            return
        if task.state == TaskState.RUNNING:
            # Can't cancel a task while it's
            # running: will raise ValueError
            # if we try, so we defer it for
            # later
            task.pending_cancellation = True
            return
        err = Cancelled()
        err.scope = task.scope
        self.throw(task, err)
        if task.state != TaskState.CANCELLED:
            # Task is stubborn, but so are we,
            # so we'll redeliver the cancellation
            # every time said task tries to call any
            # event loop primitive
            task.pending_cancellation = True
        if task.scope.done():
            task.scope.cancelled = True

    def cancel_scope(self, scope: TaskScope):
        scope.attempted_cancel = True
        for child in filter(lambda c: not c.shielded, scope.children):
            self.cancel_scope(child)
        for task in scope.tasks:
            if task is self.current_task:
                # We can't just immediately cancel the
                # current task because this method is
                # called synchronously by TaskScope.cancel(),
                # so there is nowhere to throw an exception
                # to: instead, we delay the cancellation until
                # the next tick
                if self.current_task in scope.tasks:
                    self.current_task.pending_cancellation = True
                    self.reschedule(self.current_task)
            else:
                self.cancel_task(task)
        if scope.done():
            scope.cancelled = True

    def init_pool(self, pool: TaskPool):
        pool.outer = self.current_pool
        pool.entry_point = self.current_task
        self.current_pool = pool

    def close_pool(self, pool: TaskPool):
        self.current_pool = pool.outer
        self.close_scope(pool.scope)

    def suspend(self):
        self.current_task.state = TaskState.PAUSED
        self.current_task.paused_when = self.clock.current_time()

    def setup(self):
        for manager in self.signal_managers:
            manager.install()

    def teardown(self):
        for manager in self.signal_managers:
            manager.uninstall()

    def add_shutdown_task(
        self, func: Callable[[Any, Any], Coroutine[Any, Any, Any]], *args
    ) -> int:
        ident = next(self._shutdown_task_ident)
        self.shutdown_tasks.append((ident, func, args))
        return ident

    def remove_shutdown_task(self, ident: int) -> bool:
        for task_id, func, args in self.shutdown_tasks:
            if ident == task_id:
                self.shutdown_tasks.remove((task_id, func, args))
                return True
        return False
