import warnings

import structio
from structio.abc import IOManager, Kernel
from structio.core.task import Task, TaskState
from structio.core.run import current_loop, current_task
from structio.io import FdWrapper
import select


class SimpleIOManager(IOManager):
    """
    A simple, cross-platform, select()-based
    I/O manager. This class is only meant to
    be used as a default fallback and is quite
    inefficient and slower compared to more ad-hoc
    alternatives such as epoll or kqueue (it should
    work on most platforms though)
    """

    def __init__(self):
        # Maps resources to tasks
        self.readers: dict[FdWrapper, Task] = {}
        self.writers: dict[FdWrapper, Task] = {}
        self._closed = False

    def pending(self) -> bool:
        for _, task in self.readers.items():
            if not task.done():
                return True
        for _, task in self.writers.items():
            if not task.done():
                return True
        return False

    def get_reader(self, rsc: FdWrapper):
        self._check_closed()
        return self.readers.get(rsc)

    def get_writer(self, rsc: FdWrapper):
        self._check_closed()
        return self.writers.get(rsc)

    def _collect_readers(self) -> list[int]:
        """
        Collects all resources that need to be read from,
        so we can select() on them later
        """

        return [reader.fileno() for reader in self.readers]

    def _collect_writers(self) -> list[int]:
        """
        Collects all resources that need to be written to,
        so we can select() on them later
        """

        return [writer.fileno() for writer in self.writers]

    def _check_closed(self):
        if self._closed:
            raise structio.exceptions.ResourceClosed("the I/O manager is closed")

    def wait_io(self):
        self._check_closed()
        kernel: Kernel = current_loop()
        current_time = kernel.clock.current_time()
        deadline = kernel.get_closest_deadline()
        # FIXME: This delay seems to help throttle the calls
        #   to this method. Should we be calling it this often?
        if deadline == float("inf"):
            deadline = 0.01
        elif deadline > 0:
            deadline -= current_time
        deadline = max(0, deadline)
        readers = self._collect_readers()
        writers = self._collect_writers()
        kernel.event("before_io", deadline)
        readable, writable, exceptional = select.select(
            readers,
            writers,
            writers + readers,
            deadline,
        )
        kernel.event("after_io", kernel.clock.current_time() - current_time)
        # On Windows, a successful connection is marked
        # as an exceptional event rather than a write
        # one
        writable.extend(exceptional)
        del exceptional
        for read_ready in readable:
            wrapper = FdWrapper(read_ready)
            task = self.readers[wrapper]
            kernel.reschedule(task)
            # We don't want to listen for read events on
            # this resource anymore, so we release it
            self.release(wrapper)
        for write_ready in writable:
            wrapper = FdWrapper(write_ready)
            task = self.writers[wrapper]
            kernel.reschedule(task)
            self.release(wrapper)

    def request_read(self, rsc: FdWrapper, task: Task):
        self._check_closed()
        current_task().state = TaskState.IO
        self.readers[rsc] = task

    def request_write(self, rsc: FdWrapper, task: Task):
        self._check_closed()
        current_task().state = TaskState.IO
        self.writers[rsc] = task

    def release(self, resource: FdWrapper):
        self._check_closed()
        self.readers.pop(resource, None)
        self.writers.pop(resource, None)

    def release_task(self, task: Task):
        if self._closed:
            return
        for resource, owner in self.readers.copy().items():
            if owner == task:
                self.readers.pop(resource)
        for resource, owner in self.writers.copy().items():
            if owner == task:
                self.writers.pop(resource)

    def get_readers(self) -> tuple[FdWrapper, Task]:
        self._check_closed()
        for fd, task in self.readers.copy().items():
            yield fd, task

    def get_writers(self) -> tuple[FdWrapper, Task]:
        self._check_closed()
        for fd, task in self.writers.copy().items():
            yield fd, task

    def close(self):
        self._check_closed()
        for reader in self.readers:
            warnings.warn(
                f"I/O manager was closed with scheduled write event for {reader}"
            )
        for writer in self.writers:
            warnings.warn(
                f"I/O manager was closed with scheduled write event for {writer}"
            )
        self.readers = {}
        self.writers = {}
        self._closed = True
