from structio.core.run import current_loop
from structio.abc import SchedulingPolicy
from structio.core.task import Task, TaskState
from structio.core.time.queue import TimeQueue
from collections import deque


class FIFOPolicy(SchedulingPolicy):
    """
    A First-in, First-out scheduling policy
    """

    def __init__(self):
        # Tasks that are ready to run
        self.run_queue: deque[Task] = deque()
        # Paused tasks along with their deadlines
        self.paused: TimeQueue = TimeQueue()

    # noinspection PyMethodMayBeStatic
    def is_scheduled(self, task: Task) -> bool:
        return task.state == TaskState.READY

    def has_next_task(self) -> bool:
        return bool(self.run_queue)

    def has_paused_task(self) -> bool:
        return bool(self.paused)

    def get_next_task(self) -> Task | None:
        if not self.run_queue:
            return None
        return self.run_queue.popleft()

    def peek_paused_task(self) -> Task | None:
        if not self.has_paused_task():
            return
        return self.paused.peek()

    def peek_next_task(self) -> Task | None:
        if not self.has_next_task():
            return
        return self.run_queue[0]

    def get_paused_task(self) -> Task | None:
        if not self.paused:
            return None
        return self.paused.get()[0]

    def schedule(self, task: Task):
        if self.is_scheduled(task):
            return
        task.state = TaskState.READY
        self.run_queue.append(task)

    def pause(self, task: Task):
        task.state = TaskState.PAUSED
        self.paused.put(task, task.next_deadline)

    def discard(self, task: Task):
        self.paused.discard(task)

    def get_closest_deadline(self):
        if self.has_next_task():
            # We absolutely cannot block while other
            # tasks have things to do!
            return current_loop().clock.current_time()
        deadlines = []
        for scope in current_loop().get_system_pool().scope.children:
            deadlines.append(scope.get_effective_deadline()[0])
        if not deadlines:
            deadlines.append(float("inf"))
        return min(
            [
                min(deadlines),
                self.paused.get_closest_deadline(),
            ]
        )