import inspect
import structio
import functools

# I *really* hate fork()
from multiprocessing_utils import local
from structio.abc import (
    Kernel,
    Debugger,
    Clock,
    SignalManager,
    IOManager,
    SchedulingPolicy,
)
from structio.exceptions import StructIOException
from structio.core.task import Task
from typing import Callable, Any, Coroutine

_RUN = local()


def current_loop() -> Kernel:
    """
    Returns the current event loop in the calling
    thread. Raises a StructIOException if no async
    context exists
    """
    try:
        return _RUN.kernel
    except AttributeError:
        raise StructIOException("must be called from async context") from None


def current_task() -> Task:
    """
    Shorthand for current_loop().current_task
    """

    return current_loop().current_task


def new_event_loop(kernel: Kernel):
    """
    Initializes a new event loop using the
    given kernel implementation. Cannot be
    called from an asynchronous context
    """

    try:
        current_loop()
    except StructIOException:
        _RUN.kernel = kernel
    else:
        if not _RUN.kernel.done():
            raise StructIOException(
                "cannot be called from running async context"
            ) from None
        _RUN.kernel = kernel


def run(
    func: Callable[[Any, Any], Coroutine[Any, Any, Any]],
    *args,
    kernel: type | None = None,
    policy: SchedulingPolicy| None = None,
    io_manager: IOManager | None = None,
    signal_managers: list[SignalManager] | None = None,
    clock: Clock | None = None,
    tools: list[Debugger] | None = None,
    restrict_ki_to_checkpoints: bool = False,
):
    """
    Starts the event loop from a synchronous entry point. All
    positional arguments are passed to the given coroutine
    function. If you want to pass keyword arguments, consider
    using functools.partial()
    """

    if kernel is None:
        kernel = structio.kernel.DefaultKernel
    if policy is None:
        policy = structio.core.policies.fifo.FIFOPolicy()
    if io_manager is None:
        io_manager = structio.core.managers.io.simple.SimpleIOManager()
    if clock is None:
        clock = structio.core.time.clock.DefaultClock()
    if not issubclass(kernel, Kernel):
        raise TypeError(
            f"kernel must be a subclass of {Kernel.__module__}.{Kernel.__qualname__}, not {type(kernel)}"
        )
    signal_managers = signal_managers or []
    sigint_manager = structio.core.managers.signals.sigint.SigIntManager()
    if sigint_manager not in signal_managers:
        signal_managers.append(sigint_manager)
    check = func
    if isinstance(func, functools.partial):
        check = func.func
    if inspect.iscoroutine(check):
        raise StructIOException(
            "Looks like you tried to call structio.run(your_func(arg1, arg2, ...)), that is wrong!"
            "\nWhat you wanna do, instead, is this: structio.run(your_func, arg1, arg2, ...)"
        )
    elif not inspect.iscoroutinefunction(check):
        raise StructIOException(
            "structio.run() requires an async function as its first argument!"
        )
    # Used to wake up the signal watcher when signals arrive
    waker = structio.util.wakeup_fd.WakeupFd()
    waker.set_wakeup_fd()
    new_event_loop(
        kernel(
            policy=policy,
            clock=clock,
            restrict_ki_to_checkpoints=restrict_ki_to_checkpoints,
            io_manager=io_manager,
            signal_managers=signal_managers,
            tools=tools,
        )
    )
    current_loop().spawn_system_task(structio.signals.signal_watcher, waker.reader)
    try:
        return current_loop().start(func, *args)
    finally:
        # Bunch of cleanup to ensure signals across different runs
        # do not mix up

        # noinspection PyProtectedMember
        structio.signals._sig_handlers.clear()
        # noinspection PyProtectedMember
        structio.signals._sig_data.clear()

