from enum import Enum, auto
from dataclasses import dataclass, field
from typing import Coroutine, Any, Callable
from itertools import count
import structio

_counter = count()


class TaskState(Enum):
    INIT: int = auto()
    RUNNING: int = auto()
    READY: int = auto()
    PAUSED: int = auto()
    FINISHED: int = auto()
    CRASHED: int = auto()
    CANCELLED: int = auto()
    IO: int = auto()


_joiner: Callable[[Any, Any], Coroutine[Any, Any, Any]] | None = None


@dataclass
class Task:
    """
    An asynchronous task wrapper
    """

    # The task's name
    name: str
    # The underlying coroutine of this
    # task
    coroutine: Coroutine = field(repr=False)
    # The task's scope
    scope: "TaskScope" = field(repr=False)
    # The task's pool
    pool: "TaskPool" = field(repr=False)
    # The state of the task
    state: TaskState = field(default=TaskState.INIT)
    # Used for debugging
    id: int = field(default_factory=lambda: next(_counter))
    # What error did the task raise, if any?
    exc: BaseException | None = None
    # The task's return value, if any
    result: Any | None = None
    # When did the task pause last time?
    paused_when: Any = -1
    # When is the task's next deadline?
    next_deadline: Any = -1
    # Is cancellation pending?
    pending_cancellation: bool = False
    # Is this a system task?
    is_system_task: bool = False

    def done(self):
        """
        Returns whether the task is running
        """

        return self.state in [
            TaskState.CRASHED,
            TaskState.FINISHED,
            TaskState.CANCELLED,
        ]

    def __hash__(self):
        """
        Implements hash(self)
        """

        return hash(self.coroutine)

    def cancel(self):
        """
        Cancels the given task
        """

        structio.current_loop().cancel_task(self)
