import select
import warnings
import platform
from typing import Any
import structio
from structio.abc import AsyncResource
from structio.io import FdWrapper, WantRead, WantWrite, SSLSocket
from structio.thread import run_in_worker
from structio.exceptions import ResourceClosed, ResourceBroken
from structio.core.syscalls import (
    wait_readable,
    wait_writable,
    checkpoint,
    closing,
    release,
)
from functools import wraps
import socket as _socket

try:
    import ssl as _ssl
except ImportError:
    _ssl = None


@wraps(_socket.socket)
def socket(*args, **kwargs):
    return AsyncSocket(_socket.socket(*args, **kwargs))


@wraps(_socket.fromfd)
async def fromfd(
    fd: Any,
    family: _socket.AddressFamily | int,
    type: _socket.SocketKind | int,
    proto: int = 0,
) -> "AsyncSocket":
    return AsyncSocket(_socket.fromfd(fd, family, type, proto))


async def wrap_socket_with_ssl(
    sock, *args, context, do_handshake_on_connect=True, **kwargs
):
    """
    Wraps a regular unencrypted socket or a structio async socket into a
    TLS-capable asynchronous socket. All positional and keyword arguments
    (aside from context and do_handshake_on_connect) are passed to context.wrap_socket()
    (if context is None, one with reasonable defaults is created using ssl.create_default_context()).
    Note that the do_handshake_on_connect parameter passed to the given SSL context is always False,
    because structio handles TLS handshaking on its own (this means that you mostly don't need to care
    about where do_handshake_on_connect is set: it'll just work)
    """

    if not _ssl:
        raise RuntimeError("SSL is not supported on this platform")
    if isinstance(sock, AsyncSocket):
        sock = sock.socket
    sock: _socket.socket
    context: _ssl.SSLContext
    if context is None:
        context = _ssl.create_default_context()
    # do_handshake_on_connect MUST be set to False on the
    # synchronous socket! Structio performs the TLS handshake
    # asynchronously, and letting the SSL library handle it
    # blocks the entire event loop
    raw_ssl = context.wrap_socket(sock, *args, do_handshake_on_connect=False, **kwargs)
    wrapped = AsyncSocket(raw_ssl, do_handshake_on_connect=do_handshake_on_connect)
    wrapped.socket: _ssl.SSLSocket
    if wrapped.do_handshake_on_connect and wrapped.socket._connected:
        await wrapped.do_handshake()
    return wrapped


# Wrappers of the socket module


@wraps(_socket.socketpair)
def socketpair(
    family=None, type=_socket.SOCK_STREAM, proto=0
) -> tuple["AsyncSocket", "AsyncSocket"]:
    if family is None and platform.system() == "Windows":
        family = _socket.AF_INET
    a, b = _socket.socketpair(family, type, proto)
    result = AsyncSocket(a), AsyncSocket(b)
    return result


@wraps(_socket.getaddrinfo)
async def getaddrinfo(
    host: bytearray | bytes | str | None,
    port: str | int | None,
    family: int = 0,
    type: int = 0,
    proto: int = 0,
    flags: int = 0,
):
    return await run_in_worker(
        _socket.getaddrinfo, host, port, family, type, proto, flags, cancellable=True
    )


@wraps(_socket.getfqdn)
async def getfqdn(name: str) -> str:
    return await run_in_worker(_socket.getfqdn, name, cancellable=True)


@wraps(_socket.getnameinfo)
async def getnameinfo(
    sockaddr: tuple[str, int] | tuple[str, int, int, int], flags: int
) -> tuple[str, str]:
    return await run_in_worker(_socket.getnameinfo, sockaddr, flags, cancellable=True)


@wraps(_socket.gethostname)
async def gethostname() -> str:
    return await run_in_worker(_socket.gethostname, cancellable=True)


@wraps(_socket.gethostbyaddr)
async def gethostbyaddr(ip_address: str) -> tuple[str, list[str], list[str]]:
    return await run_in_worker(_socket.gethostbyaddr, ip_address, cancellable=True)


@wraps(_socket.gethostbyname)
async def gethostbyname(hostname: str) -> str:
    return await run_in_worker(_socket.gethostbyname, hostname, cancellable=True)


@wraps(_socket.gethostbyname_ex)
async def gethostbyname_ex(hostname: str) -> tuple[str, list[str], list[str]]:
    return await run_in_worker(_socket.gethostbyname_ex, hostname, cancellable=True)


# As per RFC 8305, the default connection delay is set to
# 250 milliseconds
CONNECT_DELAY: float = 0.250


async def connect_tcp_socket(
    host: str | bytes,
    port: int,
    *,
    source_address=None,
    happy_eyeballs_delay: float = CONNECT_DELAY,
) -> "AsyncSocket":
    """
    Resolve the given (non-numeric) host and attempt to connect to it, at the chosen port.
    Connection attempts are made according to the "Happy eyeballs" algorithm as per RFC
    8305. If source_address is provided, the connection is established through that,
    otherwise we let the OS pick one. The happy_eyeballs_delay parameter controls
    how much time (in seconds) we wait for a connection attempt to stall before
    attempting the next one (by default it's set to 250ms)
    """

    # Trio states these behaviors are technically accepted by
    # (some versions of) getaddrinfo, but they're non-portable,
    # broken or not useful. And so we follow
    if host is None:
        raise ValueError("host can't be None")
    if not isinstance(port, int):
        raise TypeError(f"port must be int, got {type(port)} instead")

    hosts = await getaddrinfo(host, port, type=_socket.SOCK_STREAM)
    # RFC 8305 specifies that if we get addresses of different families,
    # our first two connection attempts should be using different ones
    # (in english: if getaddrinfo() returns, say, 2 IPV4 addresses and one IPV6
    # address, then we have to make sure our first and second attempt use one
    # of each type)
    for i in range(1, len(hosts)):
        # If the family of the ith socket (skipping
        # the first one) is different from that of
        # our very first socket (and if it isn't already
        # in second place) then we pick it and shift it
        # all the way to second place
        if hosts[i][0] != hosts[0][0] and i != 1:
            hosts.insert(1, hosts.pop(i))
            break
    if not hosts:
        # Trio is paranoid, and so are we. This should never happen
        # by the way, getaddrinfo will just raise OSError on its own
        raise OSError(f"name resolution failed for {host!r}")
    # Store all sockets we create when attempting to connect so
    # that we can shut them down later
    sockets: list[AsyncSocket] = []
    # The socket that managed to connect
    successful: AsyncSocket | None = None
    # We chain exceptions via __cause__ so that we can
    # provide information about all failed attempts if
    # we don't manage to connect
    exc_obj: Exception | None = None

    async def attempt(sock_data, addr, evt: structio.Event, scope: structio.TaskScope):
        nonlocal successful
        try:
            attempt_sock = socket(*sock_data)
            sockets.append(attempt_sock)
            if source_address:
                # This trick (again stolen from Trio), lets us
                # bind to a given address without actually busying
                # up a local port up until the moment where we actually
                # need to connect. That way, we can perform as many connection
                # attempts as we want from a given source address without ever
                # worrying about running out of local ports
                try:
                    attempt_sock.setsockopt(
                        _socket.IPPROTO_IP, _socket.IP_BIND_ADDRESS_NO_PORT, 1
                    )
                except (OSError, AttributeError):
                    # Not all platforms support this option (for example,
                    # my Linux/Windows installations don't seem to have IP_BIND_ADDRESS_NO_PORT
                    # defined in the socket module), so if setting it fails
                    # we just ignore it and hope for the best
                    pass
                # This makes sure users don't try to send IPv6
                # traffic with an IPv4 source address
                try:
                    await attempt_sock.bind((source_address, 0))
                except OSError:
                    # Almost hit the 120 character line, phew...
                    raise OSError(
                        f"Source address {source_address!r} is incompatible with remote address {addr!r}"
                    )
            await attempt_sock.connect(addr)
            # Hooray! Connection was successful. Record the socket
            # and cancel the rest of the attempts (either future or
            # currently running)
            successful = attempt_sock
            sockets.remove(attempt_sock)
            scope.cancel()
        except OSError as exc:
            # Well, this attempt failed. Right now, we just ignore the error (we'll
            # fail with OSError later if all connection attempts fail), but we should
            # really have support for ExceptionGroups (coming soon btw), so we can
            # keep track of all the errors and use fancy stuff like the new except*
            # syntax introduced in Python 3.11

            # Oh, and we also notify our next attempt that they can start early
            evt.set()

            nonlocal exc_obj
            if exc_obj:
                exc_obj.__cause__ = exc
            else:
                exc_obj = exc

    try:
        async with structio.create_pool() as pool:
            for *sock_args, _, address in hosts:
                # This event notifies us if a connection attempt
                # fails, so we can start early
                event = structio.Event()
                pool.spawn(attempt, sock_args, address, event, pool.scope)
                with structio.skip_after(happy_eyeballs_delay):
                    # We'll wait for the event to be triggered or for at
                    # most happy_eyeballs_delay seconds before moving on,
                    # whichever happens first
                    await event.wait()
    finally:
        for sock in sockets:
            if sock.fileno() == -1:
                # Socket is already dead
                continue
            try:
                # FIXME: Could this block forever? I mean, maybe it's not
                #  such a huge deal since you can always wrap the call in
                #  a timeout or something, but it may be something worth
                #  investigating
                await sock.close()
            except Exception as e:
                # Again, we shouldn't be ignoring
                # errors willy-nilly like that, but
                # hey beta software am I right?
                warnings.warn(
                    f"Failed to close {sock!r} in call to connect_tcp_socket -> {type(e).__name__}: {e}"
                )
                continue
    if not successful:
        # All connection attempts failed
        err = OSError(f"all connection attempts to {host}:{port} failed")
        if exc_obj:
            raise exc_obj from err
        raise err
    return successful


async def connect_tcp_ssl_socket(
    host: str | bytes,
    port: int,
    *,
    ssl_context=None,
    source_address=None,
    happy_eyeballs_delay: float = CONNECT_DELAY,
) -> "AsyncSocket":
    """
    Convenience wrapper over connect_socket with SSL/TLS functionality
    """

    if not _ssl:
        raise RuntimeError("SSL is not supported on the current platform")
    return await wrap_socket_with_ssl(
        await connect_tcp_socket(
            host,
            port,
            source_address=source_address,
            happy_eyeballs_delay=happy_eyeballs_delay,
        ),
        context=ssl_context,
        server_hostname=host,
    )


class AsyncSocket(AsyncResource):
    """
    Abstraction layer for asynchronous sockets
    """

    def fileno(self):
        return int(self._fd)

    def __init__(
        self,
        sock: _socket.socket,
        do_handshake_on_connect: bool = True,
    ):
        self._fd = FdWrapper(sock.fileno())
        # Do we perform the TCP handshake automatically
        # upon connection? This is only needed for SSL
        # sockets
        self.do_handshake_on_connect = do_handshake_on_connect
        self.socket = sock
        self.socket.setblocking(False)
        self.write_lock = structio.util.misc.ThereCanBeOnlyOne(
            "another task is writing on this socket"
        )
        self.read_lock = structio.util.misc.ThereCanBeOnlyOne(
            "another task is reading from this socket"
        )

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def receive(self, max_size: int, flags: int = 0) -> bytes:
        """
        Receives up to max_size bytes from a socket asynchronously
        """

        assert max_size >= 1, "max_size must be >= 1"
        if self._fd == -1:
            raise ResourceClosed("I/O operation on closed socket")
        with self.read_lock:
            while True:
                try:
                    data = self.socket.recv(max_size, flags)
                    await checkpoint()
                    return data
                except WantRead:
                    await wait_readable(self._fd)

    async def receive_exactly(self, size: int, flags: int = 0) -> bytes:
        """
        Receives exactly size bytes from a socket asynchronously
        """

        # https://stackoverflow.com/questions/55825905/how-can-i-reliably-read-exactly-n-bytes-from-a-tcp-socket
        buf = bytearray(size)
        pos = 0
        while pos < size:
            n = await self.recv_into(memoryview(buf)[pos:], flags=flags)
            if n == 0:
                raise ResourceBroken(
                    "incomplete read detected: is the remote end gone?"
                )
            pos += n
        return bytes(buf)

    async def connect(self, address):
        """
        Wrapper socket method
        """

        if self._fd == -1:
            raise ResourceClosed("I/O operation on closed socket")
        await checkpoint()
        with self.write_lock, self.read_lock:
            try:
                self.socket.connect(address)
            except WantWrite:
                await wait_writable(self._fd)
        if self.do_handshake_on_connect and hasattr(self.socket, "do_handshake"):
            await self.do_handshake()

    async def close(self):
        """
        Wrapper socket method
        """

        if self._fd.fileno() != -1:
            # We set our own fd to -1
            # before calling any async
            # primitive so that any further
            # I/O operations on this object
            # fail before they even start.
            # Of course, we still want to
            # release the actual fd, so we
            # save it separately
            fd = self._fd
            self._fd = FdWrapper(-1)
            await closing(fd)
            await release(fd)
            self.socket.close()
        else:
            await checkpoint()

    async def accept(self):
        """
        Accepts the socket, completing the 3-step TCP handshake asynchronously
        """

        if self._fd == -1:
            raise ResourceClosed("I/O operation on closed socket")
        with self.read_lock:
            while True:
                try:
                    remote, addr = self.socket.accept()
                    await checkpoint()
                    return type(self)(remote), addr
                except WantRead:
                    await wait_readable(self._fd)

    async def send_all(self, data: bytes, flags: int = 0):
        """
        Sends all the provided data asynchronously
        """

        if self._fd == -1:
            raise ResourceClosed("I/O operation on closed socket")
        with self.write_lock:
            sent_no = 0
            while data:
                try:
                    sent_no = self.socket.send(data, flags)
                    await checkpoint()
                except WantRead:
                    await wait_readable(self._fd)
                except WantWrite:
                    await wait_writable(self._fd)
                data = data[sent_no:]

    async def shutdown(self, how):
        """
        Wrapper socket method
        """

        if self.fileno() == -1:
            raise ResourceClosed("I/O operation on closed socket")
        if self.socket:
            self.socket.shutdown(how)
        await checkpoint()

    async def bind(self, addr: tuple):
        """
        Binds the socket to an address
        :param addr: The address, port tuple to bind to
        :type addr: tuple
        """

        if self._fd == -1:
            raise ResourceClosed("I/O operation on closed socket")
        self.socket.bind(addr)
        await checkpoint()

    async def listen(self, backlog: int):
        """
        Starts listening with the given backlog
        :param backlog: The socket's backlog
        :type backlog: int
        """

        if self._fd == -1:
            raise ResourceClosed("I/O operation on closed socket")
        self.socket.listen(backlog)
        await checkpoint()

    def settimeout(self, seconds):
        """
        Wrapper socket method
        """

        raise RuntimeError("Use with_timeout() to set a timeout")

    def gettimeout(self):
        """
        Wrapper socket method
        """

        return None

    def dup(self):
        """
        Wrapper socket method
        """

        return type(self)(self.socket.dup(), self.do_handshake_on_connect)

    def setsockopt(self, *args, **kwargs):
        """
        Wrapper socket method
        """

        return self.socket.setsockopt(*args, **kwargs)

    async def do_handshake(self):
        """
        Wrapper socket method
        """

        if not hasattr(self.socket, "do_handshake"):
            # Regular sockets don't have a do_handshake method
            return
        with self.read_lock, self.write_lock:
            while True:
                try:
                    self.socket: SSLSocket  # Silences pycharm warnings
                    self.socket.do_handshake()
                    await checkpoint()
                    break
                except WantRead:
                    await wait_readable(self._fd)
                except WantWrite:
                    await wait_writable(self._fd)

    async def recvfrom(self, buffersize, flags=0):
        """
        Wrapper socket method
        """

        with self.read_lock:
            while True:
                try:
                    data = self.socket.recvfrom(buffersize, flags)
                    await checkpoint()
                    return data
                except WantRead:
                    await wait_readable(self._fd)
                except WantWrite:
                    await wait_writable(self._fd)

    async def recv_into(self, buffer, nbytes=0, flags=0):
        """
        Wrapper socket method
        """

        with self.read_lock:
            while True:
                try:
                    data = self.socket.recv_into(buffer, nbytes, flags)
                    await checkpoint()
                    return data
                except WantRead:
                    await wait_readable(self._fd)
                except WantWrite:
                    await wait_writable(self._fd)

    async def recvfrom_into(self, buffer, bytes=0, flags=0):
        """
        Wrapper socket method
        """

        with self.read_lock:
            while True:
                try:
                    data = self.socket.recvfrom_into(buffer, bytes, flags)
                    await checkpoint()
                    return data
                except WantRead:
                    await wait_readable(self._fd)

    async def sendto(self, bytes, flags_or_address, address=None):
        """
        Wrapper socket method
        """

        if address:
            flags = flags_or_address
        else:
            address = flags_or_address
            flags = 0
        with self.write_lock:
            while True:
                try:
                    data = self.socket.sendto(bytes, flags, address)
                    await checkpoint()
                    return data
                except WantWrite:
                    await wait_writable(self._fd)

    def getpeername(self):
        """
        Wrapper socket method
        """

        return self.socket.getpeername()

    def getsockname(self):
        """
        Wrapper socket method
        """

        return self.socket.getsockname()

    async def recvmsg(self, bufsize, ancbufsize=0, flags=0):
        """
        Wrapper socket method
        """

        with self.read_lock:
            while True:
                try:
                    data = self.socket.recvmsg(bufsize, ancbufsize, flags)
                    await checkpoint()
                    return data
                except WantRead:
                    await wait_readable(self._fd)

    async def recvmsg_into(self, buffers, ancbufsize=0, flags=0):
        """
        Wrapper socket method
        """

        with self.read_lock:
            while True:
                try:
                    data = self.socket.recvmsg_into(buffers, ancbufsize, flags)
                    await checkpoint()
                    return data
                except WantRead:
                    await wait_readable(self._fd)

    async def sendmsg(self, buffers, ancdata=(), flags=0, address=None):
        """
        Wrapper socket method
        """

        with self.write_lock:
            while True:
                try:
                    data = self.socket.sendmsg(buffers, ancdata, flags, address)
                    await checkpoint()
                    return data
                except WantWrite:
                    await wait_writable(self._fd)

    def is_readable(self):
        """
        Returns whether the OS thinks this socket is
        readable. For more info see https://github.com/python-trio/trio/issues/760
        """

        match platform.system():
            case "Windows":
                readable, _, _ = select.select([self.socket], [], [], 0)
                return bool(readable)
            case _:
                p = select.poll()
                p.register(self.socket, select.POLLIN)
                return bool(p.poll(0))

    def __repr__(self):
        return f"AsyncSocket({self.socket})"
