"""
Helper module to spawn asynchronous Python processes via
structio
"""

import os
import sys
import struct
import msgpack
import structio
import importlib
from structio.util.finder import ObjectReference
from structio.exceptions import ResourceBroken


async def send_message(sock: structio.AsyncSocket, payload: dict):
    """
    Sends a payload to the remote event loop
    """

    data = msgpack.dumps(payload)
    await sock.send_all(struct.pack("Q", len(data)))
    await sock.send_all(data)


async def receive_message(sock: structio.AsyncSocket) -> dict:
    """
    Wait for a message from the control process
    """

    data = await sock.receive(8)
    if not data:
        raise ResourceBroken("socket was closed unexpectedly by remote loop")
    size, *_ = struct.unpack("Q", data)
    return msgpack.unpackb(await sock.receive_exactly(size))


async def send_ack(sock: structio.AsyncSocket):
    """
    Sends an acknowledgement packet back to the remote
    event loop on the given socket
    """
    return await send_message(sock, {"ok": True, "msg": "ACK"})


async def dispatch(sock: structio.AsyncSocket, message: dict):
    """
    Dispatches commands received from the remote event loop

    :param sock: The socket used to communicate with the remote
        event loop
    :param message: The message received from the remote event
        loop
    """

    match message["msg"]:
        case "HELO":
            # SOS: Start of session
            await send_ack(sock)
        case "EXEC":
            await send_ack(sock)
            if message["file"]:
                sys.path.append(os.path.dirname(message["file"]))
            obj = ObjectReference.load_ref(message["ref"])
            obj()
        case "CYA":
            # EOS: End of session (aka the process can and should exit)
            await send_ack(sock)
            sys.exit(0)
        case _:
            # IDK: I don't know (means the command is unknown)
            await send_message(
                sock,
                {
                    "ok": False,
                    "msg": "IDK",
                    "error": f"unknown message type {message['msg']!r}",
                },
            )


async def main(addr: tuple[str, int]):
    socket = structio.socket.socket()
    # Connect to the remote event loop
    await socket.connect(addr)
    async with socket:
        while True:
            await dispatch(socket, await receive_message(socket))


if __name__ == "__main__":
    structio.run(main, (sys.argv[1], int(sys.argv[2])))
