"""
Utility module to look up objects to be called in a Python subprocess spawned
by structio. Inspired by https://pikers.dev/goodboy/tractor/src/branch/mv_to_new_trio_py3.11/tractor/msg/ptr.py#L53
"""

import importlib
from pathlib import Path
from inspect import ismethod, isfunction, isbuiltin, getmodule, getmodulename
from pkgutil import resolve_name


def islambda(v) -> bool:
    """
    Returns whether the passed object is a lambda
    """

    f = lambda: ...
    return isinstance(v, type(f)) and v.__name__ == f.__name__


def get_real_module_name(obj) -> tuple[bool, str]:
    in_package = True
    module = obj.__module__

    if "__main__" in module:
        # get parent modules of object
        mod_obj = getmodule(obj)

        # from the filename of the module, get its name
        mod_suffix = getmodulename(mod_obj.__file__)

        # join parent to child with a .
        module = ".".join(filter(bool, [mod_obj.__package__, mod_suffix]))

        if mod_obj.__package__ is None:
            in_package = False

    return in_package, module


class ObjectReference:
    """
    A reference to some arbitrary, named Python object located
    in a module we have access to. This class serves the purpose
    of serializing callables into unique string identifiers that
    can be passed along to remote Python processes. This solution
    is a lot cleaner than trying to serialize objects to bytes and
    is still useful for the majority of cases, however the approach
    does have limitations, namely:
        - lambdas aren't supported, since they have no name and would
          need to be copied/serialized (which is against the intended
          design)
        - bound methods aren't supported either, because we'd have to copy
          the instance they're bound to over to the remote process. If you
          need to call instance methods in the remote process, pass a function
          that constructs the object and calls the method instead
    """

    def __init__(self, obj):
        if ismethod(obj) or islambda(obj):
            raise ValueError(
                "bound methods and lambdas cannot be passed to a remote process"
            )
        self.obj = obj
        self.in_package = True
        self._make_ref()

    def _make_ref(self) -> tuple[str, str]:
        """
        Returns a tuple (module, name) that uniquely
        represents the wrapped object `name` in its
        `module`
        """

        name: str
        if isfunction(self.obj) or isbuiltin(self.obj):
            name = self.obj.__name__
        else:
            name = type(self.obj).__name__
        in_package, module = get_real_module_name(self.obj)
        self.in_package = in_package
        return module, name

    def get_file(self):
        return str(Path(getmodule(self.obj).__file__).as_posix())

    def __str__(self):
        return ":".join(self._make_ref())

    @staticmethod
    def load_ref(ident: str):
        return resolve_name(ident)
